/* Assembler, originally by Bruce Jacob from U of MD */
/* from ENEE350H at U of Maryland, 1998              */
/* various changes made by Vince Weaver              */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "io.h"
#include "symbol.h"
#include "loader.h"
#include "dislib.h"
#include "version.h"

/* globals */
static int line;
static int text_size=0,text_offset=0;
static int bss_size=0,bss_offset=0;

	/* read and parse a line
	note that lineString must point to allocated memory,
		so that *labelPtr, *opcodePtr, and *argXPtr
		won't be pointing to readAndParse's memory
	also note that *labelPtr, *opcodePtr, and *argXPtr
		only point to memory in lineString.
	When lineString changes, so will *labelPtr,
		*opcodePtr, and *argXPtr
	returns NULL if at end-of-file */
static char *readAndParse(FILE *inFilePtr, char *lineString,
	char **labelPtr, char **opcodePtr, char **arg0Ptr,
	char **arg1Ptr, char **arg2Ptr) {
   

  char *statusString, *firsttoken;
  
   statusString = fgets(lineString, MAXLINELENGTH, inFilePtr);
   line++;
   if (statusString != NULL) {
      firsttoken = strtok(lineString, " \t\n");
      if (firsttoken == NULL || firsttoken[0] == '#') {
	    /* really?  recursion? */
	 return readAndParse(inFilePtr, lineString, labelPtr, opcodePtr, 
			     arg0Ptr, arg1Ptr, arg2Ptr);
      } else if (firsttoken[strlen(firsttoken) - 1] == ':') {
	 *labelPtr = firsttoken;
	 *opcodePtr = strtok(NULL, " \t\n");
	 firsttoken[strlen(firsttoken) - 1] = '\0';
      } else {
	 *labelPtr = NULL;
	 *opcodePtr = firsttoken;
      }
      
      if ( (*opcodePtr!=NULL) && 
	   ( (!strncmp(*opcodePtr,".byte",5)) || 
	     (!strncmp(*opcodePtr,".ascii",6)))
	   ) {
	 *arg0Ptr = strtok(NULL, "\n");
      }
      else {
         *arg0Ptr = strtok(NULL, ", \t\n");
         *arg1Ptr = strtok(NULL, ", \t\n");
         *arg2Ptr = strtok(NULL, ", \t\n");      
      }
   }
   return statusString;
}

static int isNumber(char *string) {
   
     /* return 1 if string is a number */
   int i;
   return( (sscanf(string, "%d", &i)) == 1);
}


static short reg(char *s) {
   
    int reg = atoi(s);

    if (!isNumber(s)) {
       if (s[0] == 'r' && isdigit(s[1])) {
	  reg = atoi(s+1);
       } else {
          fprintf(stderr, "error: [%s] must be a register value\n", s);
	  exit(0);
       }
    }
	
    if ((reg < 0) || (reg > 7)) {
       fprintf(stderr, "error: register value [%s/%hd] out of range\n", 
	       s, reg);
       exit(0);
    }
    return (short)(reg & 0x7);
}

static short imm(char *s) {
   
   short imm;

      /* if s is symbolic, then translate into an address */
   if (isNumber(s)) {
      imm = strtol(s,NULL,0);
   } else {
      imm = get_symbol_address(s,1,"imm",line,text_offset,bss_offset);
   }
   if ((imm < -64) || (imm > 63)) {
      fprintf(stderr, "ERROR line %d: imm offset %hd of %s out of range\n", 
	      line,imm,s);
      exit(1);
   }

   return imm & 0x7f;
}

static short raw(char *s) {

   short imm;

      /* if s is symbolic, then translate into an address */
   if (isNumber(s)) {
      imm = strtol(s,NULL,0);
   } else {
      imm = get_symbol_address(s,1,"raw",line,text_offset,bss_offset);
   }

   return imm;
}

static void usage(char *exec_name,int show_help) {
   
   fprintf(stderr,"\n%s version %s\n\n",exec_name,VERSION);
   
   if (show_help) {
      fprintf(stderr, "\tUsage: %s <assembly-code-file> "
	      "[-D] -o <machine-code-file> \n\n", exec_name);
      fprintf(stderr, "\t-D\tenables old ascii file format\n");
      fprintf(stderr, "\n");
   }
   exit(1);   
}


static FILE *object_file; 
static int old_output=0;
static int byte_offset=0;
static int byte_count=0;
static unsigned char byte_buffer[2];

static void dump_byte_buffer(unsigned char buffer[2]) {
  
   short num;
   
   num=(buffer[0]<<8)+buffer[1];
   
   if (old_output) {
      fprintf(object_file, "%04hx\n", num);
   }
   else {
      write_be_short(object_file,num);  
   }
   
//   printf("%d/%d ",buffer[0],buffer[1]);
}

static int flush_byte(int count_only) {
   
   int temp_count;
   
   if (byte_offset!=0) {
      byte_buffer[1]=0;
      byte_count++;
      if (!count_only) dump_byte_buffer(byte_buffer);
   }
   
   temp_count=byte_count;   
   byte_count=0;
   byte_offset=0;
   return temp_count;
}

static void handle_byte(char *arg0,int count_only) {
   
   char *result;
   int byte;

   result=strtok(arg0,","); 
   while(result) {
      byte=strtoul(result,NULL,0);
      byte_buffer[byte_offset]=byte;
      byte_offset++;
      byte_count++;
      if (byte_offset==2) {
	 byte_offset=0;
	 if (!count_only) dump_byte_buffer(byte_buffer);
      }
      result=strtok(NULL,",");	    	    
   }

}

static int handle_ascii(char *arg0,int count_only) {
   
   char *pointer;
   int length=0;
   unsigned char out_char=0;
   int half_offset=0;
   unsigned char half_buffer[2];
   
   pointer=arg0;
   
     /* find leading char */
   while(*arg0 && *arg0!='\"') arg0++;
   
   if (!*arg0) {
      fprintf(stderr,"ERROR: Missing quote in .ascii line %d!\n",line);
      exit(1);
   }
   
   arg0++;
   
   while(1) {
      switch(*arg0) {

       case 0: fprintf(stderr,"ERROR: missing close quote line %d!\n",line);
	       exit(1);
       case '\"': break;
       case '\\':
	 arg0++;
	 if (*arg0=='n') out_char='\n';
	 else if (*arg0=='r') out_char='\r';
	 else if (*arg0=='t') out_char='\t';
	 else if (*arg0=='\\') out_char='\\';
	 else if (*arg0=='\"') out_char='\"';
	 else if (*arg0=='\'') out_char='\'';
	 else if ((*arg0>='0') && (*arg0<='9')) {
	    if (!((*(arg0+1)>='0') && (*(arg0+1)<='9'))) {
	       out_char=0;
	       //printf("Out char is %d\n",out_char);
	    }
	    else {
	       out_char=(*arg0)-0x30;
	       arg0++;
	       if ((*arg0>='0') && (*arg0<='9')) {
		  out_char*=8;
		  out_char+=((*arg0)-0x30);
		  arg0++;
		  if ((*arg0>='0') && (*arg0<='9')) {
		     out_char*=8;
		     out_char+=((*arg0)-0x30);
		  }
	       }
	       //printf("Out char is %d\n",out_char);	  
	    }
	 }
	 else {
	   fprintf(stderr,"ERROR!  Unknown escape char line %d\n",line);
	   exit(1);
	 }
	 break;
       default:
	 out_char=*arg0;
      }
      if (*arg0=='\"') break;
      

      half_buffer[half_offset]=out_char;
      half_offset++;
      if (half_offset==2) {
         if (!count_only) dump_byte_buffer(half_buffer);
	 half_offset=0;
      }	 
      arg0++;
      length++;      
   }
   
   if (half_offset==1) {
      half_buffer[1]=0;
      if (!count_only) dump_byte_buffer(half_buffer);
      length++;
   }
   
//   printf("\nlength %d\n",length);

   return length;
}

int main(int argc, char **argv) {
   
   char *source_filename=NULL,*object_file_filename=NULL;
   char *include_filename;
   
   int backup_line=0;
   FILE *source;
   FILE *backup_source=NULL;
   short address;
   char *label, *opcode, *arg0, *arg1, *arg2;
   char lineString[MAXLINELENGTH+1];
   int i,op;
   short num=0;
   short immediateValue;
   int debug_symbols=1;
   int skip_print,in_include=0;

   
   /************************************************/
   /* Parse Command Line Arguments                 */
   /************************************************/
   for(i=0;i<argc;i++) {   

      if (argv[i][0]=='-') {
	    /* help */
	 if (argv[i][1]=='h') {
	    usage(argv[0],1);
	 }
	    /* version */
	 if (argv[i][1]=='v') {
	    usage(argv[0],0);
	 }
	    /* debug symbos */
	 if (argv[i][1]=='g') {
	    debug_symbols=1;
	 }	 
	 
	    /* old output */
	 if (argv[i][1]=='D') {
	    old_output=1;
	 }	 

	    /* output filename */
	 if (argv[i][1]=='o') {
	    if ((i+1)<argc) {
	       object_file_filename = strdup(argv[i+1]);
	       i++;
	    }
	 }
      }
      else {
         source_filename = strdup(argv[i]);
      }
   }
   
   if ((source_filename==NULL) || (object_file_filename==NULL)) {
      usage(argv[0],1);
   }


   /* open files */

   source = fopen(source_filename, "r");
   if (source == NULL) {
      fprintf(stderr, "ERROR opening source file: %s\n", source_filename);
      exit(1);
   }
   
   object_file = fopen(object_file_filename, "w");
   if (object_file == NULL) {
      fprintf(stderr, "ERROR opening object_file file: %s\n", 
	      object_file_filename);
      exit(1);
   }

   /****************************************/
   /* PASS ONE -- map symbols to addresses */
   /****************************************/

   /* assume loaded at address 0 */
   address = 0;
   line=0;

keep_going_1:   
   while(readAndParse(source, lineString, 
		      &label, &opcode, &arg0, &arg1, &arg2) != NULL) {
      
      op=get_opcode(opcode);

      if ((byte_offset!=0) && (op!=BYTE)) {
	 address+=(flush_byte(1)/2);
      }      
      
      if (label != NULL) {
	 /* look for duplicate label */
	 if (get_symbol_address(label,0,"none",line,0,0) < 0) {
	    /* label not found -- a good sign */
	    /* allocate a new symbol */
	    add_symbol(label,address,SYMBOL_TEXT);
	 } else {
	    /* duplicate label -- terminate */
	    fprintf(stderr, "ERROR line %d: duplicate label %s \n", 
		    line,label);
	    exit(1);
	 }
      }

      switch(op) {
       case MOVI:
         /* handle instructions that are more than 2 bytes */
	 address+=2;
	 break;
       case SPACE:
	 if (isNumber(arg0)) {
	    if (num_symbols>0) {
	       if (symbol[num_symbols-1].address==address) {
		  symbol[num_symbols-1].address=bss_size;
		  symbol[num_symbols-1].type=SYMBOL_BSS;
	       }
	    }
	    bss_size += (strtol(arg0,NULL,0));	    
	 } else {
	    fprintf(stderr, "error: .space needs integer argument\n");
	    exit(1);
	 }
	 break;
       case INCLUDE:
//	 fprintf(stderr,"Including file: %s\n",arg0);
	   /* save. */
	 
	 if (in_include) {
	    fprintf(stderr,"ERROR!  We don't allow nested includes\n");
	    exit(1);
	 }
	 
	 backup_source=source;
	 backup_line=line;
	 in_include=1;

	 line=0;
	 include_filename=arg0+1; /* skip quote */
	 include_filename[strlen(include_filename)-1]=0; /* drop end quote */
	 source=fopen(include_filename,"r");
	 if (source==NULL) {
	    fprintf(stderr,"Error including %s\n",include_filename);
	 }	 
	 break;
       case EQU:
//	 fprintf(stderr,"EQU %s %s\n",arg0,arg1);
	 add_symbol(arg0,strtol(arg1,NULL,0),SYMBOL_EQU);
	 break;
       case BYTE:
	 handle_byte(arg0,1);
	 break;
       case ASCII:
	 address+=handle_ascii(arg0,1)/2;
	 break;
       case INVALID:
	 break;
       default:
	 address++;
	 break;
      }
   }

   if (in_include) {
	   /* restore */
      fclose(source);
      source=backup_source;
      line=backup_line;
      in_include=0;
      goto keep_going_1;
   }
   
   address+=flush_byte(1)/2;
      
   text_offset=0;
   text_size=address;
   
   bss_offset=text_size;
   
//   dump_symbols();
   
   /******************************/
   /* PASS TWO -- ASSEMBLE       */
   /******************************/
   
   if (!old_output) {
      
      if (debug_symbols) {
         write_header(object_file,
		    get_symbol_table_size(),text_size*2,
		    text_size,text_offset,
		    bss_size,bss_offset);
      }
      else {
	 write_header(object_file,
		    0,0,
		    text_size,text_offset,
		    bss_size,bss_offset);
      }
   }

   line=0;   
   
   
   rewind(source);
   address = 0;

keep_going:
   while(readAndParse(source, lineString, &label, &opcode, 
		      &arg0, &arg1, &arg2) != NULL) {

      skip_print=0;
      
         /* this usually means label alone on line */
      if (opcode==NULL) {
	 continue;
      }
      
      op=get_opcode(opcode);

      if ((byte_offset!=0) && (op!=BYTE)) {
	 address+=flush_byte(0)/2;
      }
      
      switch(op) {
       case ADD:
	 num = (ADD<<OP_SHIFT) | 
	       (reg(arg0)<<A_SHIFT) | (reg(arg1)<<B_SHIFT) | reg(arg2);
	 break;
       case ADDI:
	 num = (ADDI<<OP_SHIFT) | 
	       (reg(arg0)<<A_SHIFT) | (reg(arg1)<<B_SHIFT) | imm(arg2);
	 break;
       case NAND:
	 num = (NAND<<OP_SHIFT) | 
	       (reg(arg0)<<A_SHIFT) | (reg(arg1)<<B_SHIFT) | reg(arg2);
	 break;
       case LUI:
	 num = (LUI<<OP_SHIFT) | 
	       (reg(arg0)<<A_SHIFT) | ((raw(arg1) >> 6) & 0x3ff);
	 break;
       case LW:
	 num = (LW<<OP_SHIFT) | 
	       (reg(arg0)<<A_SHIFT) | (reg(arg1)<<B_SHIFT) | imm(arg2);
	 break;
       case SW:
	 num = (SW<<OP_SHIFT) | 
	       (reg(arg0)<<A_SHIFT) | (reg(arg1)<<B_SHIFT) | imm(arg2);
	 break;
       case JALR:
	 num = (JALR<<OP_SHIFT) | 
	       (reg(arg0)<<A_SHIFT) | (reg(arg1)<<B_SHIFT);
	 break;
       case BEQ:
	 /* if arg2 is symbolic, then translate into an address */
	 if (isNumber(arg2)) {
	    immediateValue = strtol(arg2,NULL,0);
	 } else {
	    immediateValue =get_symbol_address(arg2,1,"beq",line,
					       text_offset,bss_offset)
	      -address-1;
	 }
	 
	 if (immediateValue < -64 || immediateValue > 63) {
	    fprintf(stderr, "ERROR line %d: beq offset %hd out of range\n", 
		   line,immediateValue);
	    exit(1);
	 }
	 
	 num = (BEQ<<OP_SHIFT) | 
	       (reg(arg0)<<A_SHIFT) | (reg(arg1)<<B_SHIFT) | 
	       (immediateValue & 0x7f);
         break;
	 
       case NOP:
	 num = (ADD<<OP_SHIFT) | 
	       (reg("0") << A_SHIFT) | (reg("0") << B_SHIFT) | reg("0");
	 break;
       case HALT:
	 num = (EXT<<OP_SHIFT) | 
	       (reg("0") << A_SHIFT) | (reg("0") << B_SHIFT) | 
	       (EXT_EXCEPTION << 4) | EXC_HALT;
	 break;
       case SYS:
	 num = (EXT << OP_SHIFT) | 
	       (reg("0") << A_SHIFT) | (reg("0") << B_SHIFT) | 
	       (EXT_SYSCALL << 4) | (raw(arg0) & 0xf);
	 break;
       case EXC:
	 num = (EXT<<OP_SHIFT) | 
	       (reg("0") << A_SHIFT) | (reg("0") << B_SHIFT) | 
	       (EXT_EXCEPTION << 4) | (raw(arg0) & 0xf);
	 break;
       case LLI:
	 num = (ADDI << OP_SHIFT) | 
	       (reg(arg0) << A_SHIFT) | (reg(arg0) << B_SHIFT) | 
	       (raw(arg1) & 0x3f);
         break;
       case MOVI:
	 num = (LUI << OP_SHIFT) | 
	       (reg(arg0) << A_SHIFT) | ((raw(arg1) >> 6) & 0x3ff);
	 if (old_output) {
	    fprintf(object_file, "%04hx\n", num);
	 }
	 else {
	    write_be_short(object_file,num);  
	 }
	 address++;
	 num = (ADDI << OP_SHIFT) | 
	       (reg(arg0)<<A_SHIFT) | (reg(arg0)<<B_SHIFT) | (raw(arg1)&0x3f);
	 break;
       case FILL:	 
	 num = raw(arg0);
	 break;
       case SPACE:	 
	/* we put bss at end */
	 skip_print=1;
	 break;
       case INCLUDE:
	 
	 fprintf(stderr,"Including file: %s\n",arg0);
	   /* save. */
	 
	 if (in_include) {
	    fprintf(stderr,"ERROR!  We don't allow nested includes\n");
	    exit(1);
	 }
	 
	 backup_source=source;
	 backup_line=line;
	 in_include=1;

	 line=0;
	 include_filename=arg0+1; /* skip quote */
	 include_filename[strlen(include_filename)-1]=0; /* drop end quote */	 
	 source=fopen(include_filename,"r");
	 if (source==NULL) {
	    fprintf(stderr,"Error including %s\n",include_filename);
	 }
	 
	 skip_print=1;
	 break;
       case EQU:
	 skip_print=1;
	 break;
       case BYTE:
//	 fprintf(stderr,"BYTE %s\n",arg0);
	 handle_byte(arg0,0);
	 skip_print=1;
	 break;
       case ASCII:
	 address+=handle_ascii(arg0,0)/2;
	 skip_print=1;
	 break;
       case INVALID:	 
       default:
        fprintf(stderr, "error: unrecognized opcode [%s] at line %d\n", 
		opcode, line);
	exit(1);
	break;
      }

          /* print the value.  Don't do it if bss command */
       if (!skip_print) {
          if (old_output) {
	     fprintf(object_file,"%04hx\n",num);
          }
          else {
             write_be_short(object_file,num);
          }
          address++;       
       }

   }
   
   if (in_include) {
	   /* restore */
      fclose(source);
      source=backup_source;
      line=backup_line;
      in_include=0;
      goto keep_going;
   }

   address+=flush_byte(0)/2;  /* flush any pending odd-numbered .byte directives */
   
   /* handle bss segment */
   for(i=0;i<bss_size;i++) {
      if (old_output) { 
	 fprintf(object_file, "%04hx\n", 0);
      }
   }

      /* dump symbols */
   if ((!old_output) && (debug_symbols)) {
//      printf("Writing %d symbols at %ld (%lx)\n",
//	     num_symbols,ftell(object_file),ftell(object_file));
      dump_symbols_to_disk(object_file);

   }
   
   fclose(source);   
   fclose(object_file);
   return 0;
}

