/* ********************************************************************** *\
\* The Mysterious LC-998 Simulator!                                       */
/*     Re-written from scratch by Vince Weaver (weave _at_ eng.umd.edu)   *\
\*        For ENEE350H, Fall 1998.             (vince _at_ deater.net)    */
/*            Version 0.0.1, 20 September 1998                            *\
\*            Version 0.0.2,  4 September 2009                            */
/*               Originally written under Linux 2.1.121, ported to Solaris*\
\* ********************************************************************** */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

/* for syscall emulation */
#include <sys/utsname.h> /* uname() */
#include <fcntl.h>       /* open()  */

#include "dislib.h"
#include "version.h"
#include "loader.h"


#define MEMSIZE 0x8000        /* 64K, those were the days */

short system_ram[MEMSIZE];
short PC;
short reg[8];

    /* Display Help and/or Version Info */
void display_help(char *executable,int show_version_only,int extra_help) {
    printf("\nThe LC-998 Simulator version %s\n",VERSION);
    printf("    by Vince Weaver (weave@deater.net)\n");
    if (!show_version_only) {
       printf("\nUsage: %s input_file\n",executable);
       printf("       \"%s -h\" for extra help.\n",executable);
    }
    if (extra_help) {
       printf("\n       -h : show extra help\n");
       printf("       -v : show version info\n");
       printf("       -s : show memory dump\n");
       printf("       -t : enable trace info\n");
    }
    printf("\n");
}


    /* Print the CPU State */
void print_cpustate(int clock_cycles,int program_size,
		    short assembler,short pc,
		    int supress_memdump,int disable_trace) {
 
int i;
char assembler_code[BUFSIZ];
   
    if (!disable_trace) {
       printf("State After Cycle %i\n",clock_cycles);
       decode_assembly(assembler,pc,assembler_code);
       printf("   Old PC:\t%04x\n   New PC:\t%04hx\n",pc,PC);
       printf("   Command Executed:\t%04hx = %s\n",assembler,assembler_code);
       for(i=0;i<4;i++) printf("\treg[%i] %04hx",i,reg[i]);
       printf("\n");
       for(i=4;i<8;i++) printf("\treg[%i] %04hx",i,reg[i]);
       printf("\n");
       if (!supress_memdump) {
          printf("   Memory Dump:\n");
          for(i=0;i<program_size;i++) {
	     decode_assembly(system_ram[i],i,assembler_code);
	     printf("\t %04hx : %04hx %s\n",i,system_ram[i],assembler_code);
          }
       }
       printf("\n");
    }
   
}

static void copy_to_risc(char *buffer,int size,
			 short risc_addr) {
		    
   int i;
   
   for(i=0;i<size/2;i++) {
      system_ram[risc_addr+i]=((buffer[i*2]&0xff)<<8)+
			                  ((buffer[(i*2)+1]&0xff));
   }
}

static int risc_strlen(short risc_addr) {
		    
   int count=0,len=0;
   
   while(1) {
      if (((system_ram[risc_addr+count]>>8)&0xff)==0) break;
      len++;
      if ((system_ram[risc_addr+count]&0xff)==0) break;
      len++;
      count++;
   }
   return len;
}

static void copy_from_risc(char *buffer,int size,
			 short risc_addr) {
		    
   int i;
   
   for(i=0;i<size/2;i++) {
      buffer[i*2]=((system_ram[risc_addr+i]>>8)&0xff);
      buffer[(i*2)+1]=((system_ram[risc_addr+i])&0xff);
   }
      /* handle odd byte */
   if (size&1) buffer[i*2]=((system_ram[risc_addr+i]>>8)&0xff);
}



#define SYS_EXIT  1
#define SYS_READ  3
#define SYS_WRITE 4
#define SYS_OPEN  5
#define SYS_CLOSE 6
#define SYS_SYSINFO 116
#define SYS_UNAME 122

short do_syscall(short number,short arg1,short arg2,short arg3, short arg4) {
   
   int result=0,length;
   char *buffer;
   
   struct utsname uname_info;
   
   switch(number) {
      
    case SYS_OPEN:
                    length=risc_strlen(arg1);
                    buffer=malloc(length+1);
                    copy_from_risc(buffer,length,arg1);
                    buffer[length]=0;
                    result=open(buffer,arg2);
                    free(buffer);
                    break;
    case SYS_CLOSE: 
                    result=close(arg1);
                    break;
    case SYS_READ:  
                    buffer=malloc(arg3);
                    result=read(arg1,buffer,arg3);
                    copy_to_risc(buffer,result,arg2);
                    free(buffer);
                    break;
    case SYS_WRITE:
                    buffer=malloc(arg3);
                    copy_from_risc(buffer,arg3,arg2);
      		    result=write(arg1,buffer,arg3);
                    free(buffer);
		    break;
    case SYS_SYSINFO:
                    /* These are 16 bit values */
                    /* load1, load5, load 15, Total RAM */
                    /* Total RAM is in kB */
                    buffer=malloc(8);
                    buffer[0]=0; buffer[1]=0;
                    buffer[2]=0; buffer[3]=0;
                    buffer[4]=0; buffer[5]=0;
                    buffer[6]=0; buffer[7]=128;
                    copy_to_risc(buffer,8,arg1);
                    free(buffer);
                    break;
    case SYS_UNAME:
                    /* RiSC wants 5 32-byte null terminated strings */
                    /* sysname,nodename,release,version,machine */
                    uname(&uname_info);
      
                    buffer=malloc(160);
      
                    strncpy(buffer,uname_info.sysname,31);
                    buffer[31]=0;
                    copy_to_risc(buffer,32,arg1);
      
                    strncpy(buffer,uname_info.nodename,31);
                    buffer[31]=0;
                    copy_to_risc(buffer,32,arg1+16);     
      
                    strncpy(buffer,uname_info.release,31);
                    buffer[31]=0;
                    copy_to_risc(buffer,32,arg1+32);           
      
                    strncpy(buffer,uname_info.version,31);
                    buffer[31]=0;
                    copy_to_risc(buffer,32,arg1+48);           
      
                    strncpy(buffer,uname_info.machine,31);
                    buffer[31]=0;
                    copy_to_risc(buffer,32,arg1+64);      
      
                    free(buffer);
       
                    break;
    case SYS_EXIT:  
                    printf("***** EXIT(%d) SYSCALL   *****\n",arg1);
	            result=arg1;
                    break;
    default:
		    printf("ERROR!  UNIMPLEMENTED SYSCALL %d!\n",number);
                    exit(1);
   }

   return result;
}
   
    /* The Main Program */
int main(int argc,char **argv) {

    short opcode,arg1,arg2,arg3,arg4;
    int c,i,clock_cycle,program_size,result;
    extern int optind;
    short temp_short;
    unsigned short current_instruction,current_pc;
    int suppress_memdump=1,disable_trace=1;
    char input_filename[BUFSIZ];
   
       /* Parse Command Line Options */
    while ((c =getopt(argc,argv,"HhVvSsTt")) != -1)
       switch (c) {
	case 'h': case 'H': display_help(argv[0],0,1); return 0; break;
	case 'v': case 'V': display_help(NULL,1,0); return 0; break;
	case 's': case 'S': suppress_memdump=0; break;
	case 't': case 'T': disable_trace=0; break;
    }

    if (optind!=argc-1) {
       display_help(argv[0],0,0);
       printf("\n**ERROR** You must include both the name of the file\n"
	      "          which contains the machine language you wish\n"
              "          to simulate!!!\n\n");
       return 1;	  
    }
    strncpy(input_filename,argv[optind],BUFSIZ);
   
    open_object(input_filename);
       
    program_size=0;
   
       /* Load program into main RAM */   
    if (object.type==OBJECT_OLD) {
       
       while (read_short(&temp_short)>=0) {
          if (program_size>MEMSIZE) {
	     printf("\n**ERROR** Program size bigger than size of available\n"
		 "          RAM (%i bytes).\n\n",MEMSIZE);
	     return 1;
          }
          system_ram[program_size]=temp_short;
          program_size++;
       }
    }
    else {
          /* Load Text */
       for(i=0;i<object.text_size/2;i++) {
	  read_short(&temp_short);
	  system_ram[program_size]=temp_short;
	  program_size++;
	  if (program_size>MEMSIZE) {
	     fprintf(stderr,"ERROR!  Program size bigger than available "
		     "memory (%d Bytes)\n",MEMSIZE);
	     exit(1);
	  }
       }
//       printf("Loaded %d bytes up to address %04x\n",
//	      i,program_size);
       
          /* Load BSS */
       for(i=0;i<object.bss_size/2;i++) {
	  system_ram[object.bss_offset+i]=0;
       }
       
    }
    close_object();
    
       /* Initialize Processor */
    PC=0;
    clock_cycle=0;
    for(i=0;i<8;i++) reg[i]=0;
       /* Start Executing! */
    while (1) {
       current_instruction=system_ram[PC];
       current_pc=PC;
       split_instruction(current_instruction,&opcode,&arg1,&arg2,&arg3,&arg4);
       
       PC++;
       switch(opcode) {
	case ADD: /*ADD*/
	         reg[arg1]=reg[arg2]+reg[arg3];
	         reg[0]=0;
	         break;
	case ADDI: /*ADDI*/
	         reg[arg1]=reg[arg2]+arg3;
	         reg[0]=0;
	         break;
	case NAND: /*NAND*/
	         reg[arg1]=~(reg[arg2]&reg[arg3]);
	         reg[0]=0;
	         break;
	case LUI: /*LUI*/
	         reg[arg1]=(arg2<<6);
	         reg[0]=0;
	         break;
	case SW: /*SW*/
	         system_ram[reg[arg2]+arg3]=reg[arg1];
	         break;
	case LW: /*LW*/
	         if (reg[arg1]==system_ram[reg[arg2]+arg3]) {
//		    printf("r%d already equals %x at PC %x\n",
//			   arg1,reg[arg1],PC);
		 }
	         else {
	            reg[arg1]=system_ram[reg[arg2]+arg3];
		 }
	         reg[0]=0;
	         break;	
        case BEQ: /*BEQ*/
	         if (reg[arg1]==reg[arg2]) PC+=arg3;
	         break;
	case JALR: /* JALR and Extended */
	  
	            /* We have an extended opcode */
	  	 if ((arg1==0) && (arg2==0)) {
		    if (arg3==EXT_EXCEPTION) {
		       if (arg4==EXC_HALT) {
			  /* HALT */
	                  printf("***** SIMULATION HALTED *****\n");
	                  print_cpustate(clock_cycle,program_size,
				         current_instruction,current_pc,
					 suppress_memdump,0);
	                  return 0; 
		       }
		       if (arg4==EXC_DEBUG) {
			  /* Dump Registers */
			  print_cpustate(clock_cycle,program_size,
				         current_instruction,current_pc,
					 suppress_memdump,0);
		       }
		       else {
			  printf("UNKNOWN EXCEPTION %d!\n",arg4);
		       }
		    }
		    
		    if (arg3==EXT_SYSCALL) {
		       result=do_syscall(reg[1],reg[2],reg[3],reg[4],reg[5]);
		       if (reg[1]==SYS_EXIT) {
			  printf("***** SIMULATION HALTED *****\n");
	                  print_cpustate(clock_cycle,program_size,
				         current_instruction,current_pc,
					 suppress_memdump,0);
                          reg[1]=result;
	                  return result;
		       }
		       else {
			  reg[1]=result;  
		       }
		    }
		 }
	         else {
	         
		    /* JALR */
	  
	            reg[arg1]=PC;
	            PC=reg[arg2];
	            reg[0]=0;
		 }
	         break;
	
	default: break;
	  
       }

       print_cpustate(clock_cycle,program_size,
		      current_instruction,
		      current_pc,
		      suppress_memdump,disable_trace);
       clock_cycle++;
    }
    print_cpustate(clock_cycle,program_size,
		   current_instruction,
		   current_pc,suppress_memdump,0);
    return 0;  
}
